import json
from typing import Any, TYPE_CHECKING
from ..context.context import Context
if TYPE_CHECKING:
    from .. import dispatcher
from bclib.utility import DictEx


class RabbitContext(Context):
    """Context for rabbit-mq request"""

    def __init__(self, rabbit_message: DictEx,  dispatcher: 'dispatcher.IDispatcher'):
        super().__init__(dispatcher)
        self.__rabbit_message: DictEx = rabbit_message
        self.__message = DictEx(json.loads(
            rabbit_message.message)) if rabbit_message.message else None
        self.url = self.__rabbit_message.host

    @property
    def host(self) -> 'str':
        return self.__rabbit_message.host

    @property
    def queue(self) -> 'str':
        return self.__rabbit_message.queue

    @property
    def raw_message(self) -> 'str':
        return self.__rabbit_message.message

    @property
    def message(self) -> 'DictEx':
        return self.__message



import asyncio
import cgi
import io
import datetime
import json
import uuid
import base64
from typing import Callable, Coroutine, TYPE_CHECKING, Optional

from urllib.parse import unquote, parse_qs
from ..endpoint import Endpoint
from ..http_listener.http_base_data_name import HttpBaseDataName
from ..http_listener.http_base_data_type import HttpBaseDataType
from bclib.utility import DictEx
from ..message_type import MessageType
from ..message import Message

if TYPE_CHECKING:
    from aiohttp import web

from aiohttp.log import web_logger

class HttpListener:
    _id = 0
    LOGGER = "logger"
    ROUTER = "router"
    MIDDLEWARES = "middlewares"
    HANDLER_ARGS = "handler_args"
    CLIENT_MAX_SIZE = "client_max_size"

    _DEFAULT_LOGGER = web_logger
    _DEFAULT_ROUTER = None
    _DEFAULT_MIDDLEWARES = ()
    _DEFAULT_HANDLER_ARGS = None
    _DEFAULT_CLIENT_MAX_SIZE = 1024 ** 2
    

    def __init__(self, endpoint: Endpoint, async_callback: 'Callable[[Message], Coroutine[Message]]', configuration: Optional[DictEx]):
        self.__endpoint = endpoint
        self.on_message_receive_async = async_callback
        self.__config = configuration if configuration is not None else DictEx()
        self.__logger = self.__config.get(HttpListener.LOGGER, HttpListener._DEFAULT_LOGGER)
        self.__router = self.__config.get(HttpListener.ROUTER, HttpListener._DEFAULT_ROUTER)
        self.__middlewares = self.__config.get(HttpListener.MIDDLEWARES, HttpListener._DEFAULT_MIDDLEWARES)
        self.__handler_args = self.__config.get(HttpListener.HANDLER_ARGS, HttpListener._DEFAULT_HANDLER_ARGS)
        self.__client_max_size = self.__config.get(HttpListener.CLIENT_MAX_SIZE, HttpListener._DEFAULT_CLIENT_MAX_SIZE)
        

    def initialize_task(self, event_loop: asyncio.AbstractEventLoop):
        event_loop.create_task(self.__server_task(event_loop))

    async def __server_task(self, event_loop: asyncio.AbstractEventLoop):
        from aiohttp import web
        from multidict import MultiDict

        async def on_request_receive_async(request: 'web.Request') -> web.Response:
            ret_val: web.Response = None
            cms_object = await self.create_cms_async(request)
            msg = Message(str(uuid.uuid4()), MessageType.AD_HOC,
                          json.dumps(cms_object).encode())
            result = await self.on_message_receive_async(msg)
            if result:
                cms: dict = json.loads(result.buffer.decode("utf-8"))
                header_code: str = cms[HttpBaseDataType.CMS][HttpBaseDataType.WEB_SERVER]["headercode"]
                mime = cms[HttpBaseDataType.CMS][HttpBaseDataName.WEB_SERVER]["mime"]
                headers: MultiDict = None
                if HttpBaseDataName.HTTP in cms[HttpBaseDataType.CMS]:
                    http: dict = cms[HttpBaseDataType.CMS][HttpBaseDataName.HTTP]
                    if http:
                        headers = MultiDict()
                        for key, value in http.items():
                            if isinstance(value, list):
                                for item in value:
                                    headers.add(key, item)
                            else:
                                headers.add(key, value)
                ret_val = web.Response(
                    status=int(header_code.split(' ')[0]),
                    headers=headers,
                    content_type=mime)
                if HttpBaseDataName.CONTENT in cms[HttpBaseDataType.CMS]:
                    ret_val.text = cms[HttpBaseDataType.CMS][HttpBaseDataName.CONTENT]
                else:
                    raw_blob_content = cms[HttpBaseDataType.CMS][HttpBaseDataName.BLOB_CONTENT]
                    ret_val.body = base64.b64decode(raw_blob_content.encode("utf-8"))
            else:
                ret_val = web.Response()
            return ret_val

        app = web.Application(
            logger=self.__logger,
            router=self.__router,
            middlewares=self.__middlewares,
            handler_args=self.__handler_args,
            client_max_size=self.__client_max_size,
            loop=event_loop
        )
        app.add_routes(
            [web.route('*', '/{tail:.*}', on_request_receive_async)])
        runner = web.AppRunner(app, handle_signals=True)
        await runner.setup()
        site = web.TCPSite(runner, self.__endpoint.url, self.__endpoint.port)
        await site.start()
        print(
            f"Development Edge server started at http://{self.__endpoint.url}:{self.__endpoint.port}")
        try:
            while True:
                await asyncio.sleep(.0001)
        except asyncio.CancelledError:
            pass
        finally:
            print("Development Edge server stopped.")
            await site.stop()
            await runner.cleanup()
            await runner.shutdown()

    @staticmethod
    async def create_cms_async(request: 'web.Request') -> dict:
        cms_object = dict()
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.METHODE, request.method.lower())
        raw_url = unquote(request.path_qs)[1:]
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.RAW_URL, raw_url)
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.URL, request.path[1:])
        HttpListener.__add_query_string(request.query, cms_object)
        for key, value in request.headers.items():
            field_name = key.strip().lower()
            if field_name == HttpBaseDataName.COOKIE:
                HttpListener.__add_cookie(value, cms_object)
            elif field_name == HttpBaseDataName.HOST:
                HttpListener.__add_host(value, raw_url, cms_object)
            else:
                HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                          field_name, str(value).strip())
        HttpListener.__add_server_data(cms_object, request)
        await HttpListener.__add_body_async(cms_object, request)
        return {"cms": cms_object}

    @staticmethod
    async def __add_body_async(cms_object: dict, request: 'web.Request'):
        content_len_str = request.headers.get('Content-Length')
        if content_len_str or request.can_read_body:
            raw_body = await request.read()
            body = raw_body.decode('utf-8')
            HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                      HttpBaseDataName.BODY, body)
            content_type: str = request.headers.get(
                "content-type")
            if content_type and content_type.find("application/json") < 0:
                if content_type.find("multipart/form-data") >= 0:
                    _, content_type_value_params = cgi.parse_header(
                        content_type)
                    content_type_value_params['boundary'] = bytes(
                        content_type_value_params['boundary'], "utf-8")
                    if content_len_str is not None:
                        content_type_value_params['CONTENT-LENGTH'] = int(
                            content_len_str)
                    with io.BytesIO(raw_body) as stream:
                        fields = cgi.parse_multipart(
                            stream, content_type_value_params)
                    for key, value in fields.items():
                        HttpListener.__add_header(cms_object,
                                                  HttpBaseDataType.FORM, key, value[0] if len(value) == 1 else value)
                else:
                    for key, value in parse_qs(body).items():
                        HttpListener.__add_header(
                            cms_object, HttpBaseDataType.FORM, key, value[0] if len(value) == 1 else value)

    @staticmethod
    def __add_server_data(cms_object: dict, request: 'web.Request'):
        HttpListener._id += 1
        now = datetime.datetime.now()
        HttpListener.__add_header(cms_object, HttpBaseDataType.CMS,
                                  HttpBaseDataName.DATE, now.strftime("%d/%m/%Y"))
        HttpListener.__add_header(cms_object, HttpBaseDataType.CMS,
                                  HttpBaseDataName.TIME, now.strftime("%H:%M"))
        HttpListener.__add_header(cms_object, HttpBaseDataType.CMS,
                                  HttpBaseDataName.DATE2, now.strftime("%Y%m%d"))
        HttpListener.__add_header(cms_object, HttpBaseDataType.CMS,
                                  HttpBaseDataName.TIME2, now.strftime("%H%M%S"))
        HttpListener.__add_header(cms_object, HttpBaseDataType.CMS,
                                  HttpBaseDataName.DATE3, now.strftime("%Y.%m.%d"))
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.REQUEST_ID, str(HttpListener._id))
        host_parts = request.host.split(':')
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.HOST_IP, host_parts[0])
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.HOST_PORT,  host_parts[1] if len(host_parts) > 1 else "80")  # edit
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.CLIENT_IP, str(request.remote))

    @staticmethod
    def __add_query_string(query: dict, cms_object: dict) -> None:
        for key, value in query.items():
            HttpListener.__add_header(
                cms_object, HttpBaseDataType.QUERY, key, value)

    @staticmethod
    def __add_cookie(raw_header_value: str, cms_object) -> None:
        for item in raw_header_value.split(';'):
            parts = item.split('=')
            if len(parts) == 2:
                HttpListener.__add_header(cms_object, HttpBaseDataName.COOKIE,
                                          parts[0].strip(), parts[1].strip())

    @staticmethod
    def __add_host(row_host: str, row_url: str, cms_object) -> None:
        host_parts = row_host.split(':')
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.HOST, host_parts[0])
        if len(host_parts) == 2:
            HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                      HttpBaseDataName.PORT, host_parts[1])
        HttpListener.__add_header(cms_object, HttpBaseDataType.REQUEST,
                                  HttpBaseDataName.FULL_URL,  f"{row_host}/{row_url}")

    @staticmethod
    def __add_header(cms_object: dict, value_type: str, value_name: str, value: str) -> None:
        if value_type not in cms_object:
            cms_object[value_type] = dict()
        type_node = cms_object[value_type]
        if value_name in type_node:
            name_node = type_node[value_name]
            if isinstance(name_node, list):
                name_node.append(value)
            else:
                type_node[value_name] = [name_node, value]
        else:
            type_node[value_name] = value



import asyncio
from io import BufferedReader, BufferedWriter
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bclib.listener.message import Message


class LinuxNamedPipeHelper:

    @staticmethod
    def __read_from_named_pipe(named_pipe_handle: BufferedReader, future: asyncio.Future) -> None:
        from bclib.listener import Message, MessageType

        try:
            data = named_pipe_handle.read(1)
            if data:
                message_type = MessageType(int.from_bytes(
                    data, byteorder='big', signed=True))
                data = named_pipe_handle.read(4)
                data_len = int.from_bytes(
                    data, byteorder='big', signed=True)
                data = named_pipe_handle.read(data_len)
                session_id = data.decode("utf-8")
                parameter = None
                if message_type in (MessageType.AD_HOC, MessageType.MESSAGE, MessageType.CONNECT):
                    data = named_pipe_handle.read(4)
                    data_len = int.from_bytes(
                        data, byteorder='big', signed=True)
                    parameter = named_pipe_handle.read(
                        data_len)
                message = Message(session_id, message_type, parameter)
            else:
                message = None
            future.get_loop().call_soon_threadsafe(future.set_result, message)
        except Exception as ex:
            future.get_loop().call_soon_threadsafe(future.set_exception, ex)

    @staticmethod
    async def read_from_named_pipe_async(named_pipe_handle: BufferedReader, loop: asyncio.AbstractEventLoop = None) -> 'Message':
        if not loop:
            loop = asyncio.get_event_loop()
        future = loop.create_future()
        loop.run_in_executor(
            None, LinuxNamedPipeHelper.__read_from_named_pipe, named_pipe_handle, future)
        return await future

    @staticmethod
    def write_to_named_pipe(message: 'Message', named_pipe_handle: BufferedWriter):
        from bclib.listener import MessageType

        named_pipe_handle.write(message.type.value.to_bytes(1, 'big'))

        data = message.session_id.encode()
        data_length_bytes = len(data).to_bytes(4, 'big')
        named_pipe_handle.write(data_length_bytes)
        named_pipe_handle.write(data)

        if message.type in (MessageType.AD_HOC, MessageType.MESSAGE):
            data_length_bytes = len(message.buffer).to_bytes(4, 'big')
            named_pipe_handle.write(data_length_bytes)
            named_pipe_handle.write(message.buffer)

        named_pipe_handle.flush()



from ..cache.manager import CacheManager
from bclib.utility import DictEx
from ..cache.signaler.factory import SignalerFactory
import asyncio

class SignalBaseCacheManager(CacheManager):
    DEFAULT_CLEAN_INTERVAL = 43200 #Seconds => 12 Hours; 0 for indefinitely
    DEFAULT_RESET_INTERVAL = 86400 #Seconds => 24 Hours; 0 for indefinitely

    def __init__(self, options: "DictEx") -> None:
        super().__init__(options)
        self.__clean_interval = int(self._options.clean_interval) if self._options.has("clean_interval") else SignalBaseCacheManager.DEFAULT_CLEAN_INTERVAL
        if self.__clean_interval < 0:
            raise ValueError("Invalid input for clean_interval!")
        self.__reset_interval = int(self._options.reset_interval) if self._options.has("reset_interval") else SignalBaseCacheManager.DEFAULT_RESET_INTERVAL
        if self.__reset_interval < 0:
            raise ValueError("Invalid input for reset_interval!")
        signaler_options = self._options.signaler if self._options.has("signaler") else None
        self._reset_signaler = SignalerFactory.create(self.reset, signaler_options)
        if self.__reset_interval > 0 or self.__clean_interval > 0:
            loop = asyncio.get_event_loop()
            if self.__reset_interval > 0:
                loop.create_task(self.__reset_async(self.__reset_interval))
            if self.__clean_interval > 0:
                loop.create_task(self.__clean_async(self.__clean_interval))
    
    async def __reset_async(self, interval:"int"):
        try:
            while True:
                await asyncio.sleep(interval)
                self.reset()
        except: ...

    async def __clean_async(self, interval:"int"):
        try:
            while True:
                await asyncio.sleep(interval)
                self.clean()
        except: ...



from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .. import dispatcher

from bclib.parser import HtmlParserEx
from bclib.utility import DictEx
from ..context.context import Context


class ServerSourceContext(Context):
    """Base class for dispatching server base dbsource request context"""

    def __init__(self, cms_object: dict,  dispatcher: 'dispatcher.IDispatcher') -> None:
        super().__init__(dispatcher)
        parser = HtmlParserEx()
        self.raw_command = cms_object["command"]
        self.dmn_id = cms_object["dmnid"] if "dmnid" in cms_object else None
        self.params = DictEx(
            cms_object["params"]) if "params" in cms_object else None
        parser.feed(self.raw_command)
        self.command = parser.get_dict_ex()
        self.process_async = True

    def generate_response(self, result: Any) -> dict:
        return result



import json
from typing import Any, TYPE_CHECKING

from bclib.utility import DictEx, HttpStatusCodes, HttpMimeTypes, ResponseTypes
from bclib.exception import ShortCircuitErr
from bclib.context.context import Context

if TYPE_CHECKING:
    from .. import dispatcher


class RequestContext(Context):
    """Base class for dispatching http base request context"""

    def __init__(self, cms_object: dict,  dispatcher: 'dispatcher.IDispatcher') -> None:
        super().__init__(dispatcher)
        self.cms = DictEx(cms_object)
        self.url: str = self.cms.request.url
        self.query: DictEx = self.cms.query
        self.form: DictEx = self.cms.form
        self.__headers: dict = None
        self.response_type: str = ResponseTypes.RENDERED
        self.status_code: str = HttpStatusCodes.OK
        self.mime = HttpMimeTypes.HTML

    def add_header(self, key: str, value: str) -> None:
        """Adding item to response header"""

        if self.__headers is None:
            self.__headers = dict()
        if key not in self.__headers:
            self.__headers[key] = [value]
        else:
            self.__headers[key].append(value)

    def generate_error_response(self, exception: Exception) -> dict:
        """Generate error response from process result"""
        error_object, self.status_code = self._generate_error_object(exception)
        if isinstance(exception, ShortCircuitErr) and exception.data:
            content = exception.data if isinstance(
                exception.data, str) else json.dumps(exception.data, indent=1).replace("\n", "</br>")
        else:
            content = f"{error_object['errorMessage']} (Error Code: {error_object['errorCode']})"
            if 'error' in error_object:
                error = error_object["error"].replace("\n", "</br>")
                content += f"<hr/>{error}"
        self.mime = HttpMimeTypes.HTML
        return self.generate_response(content)

    def generate_response(self, content: Any) -> dict:
        """Generate response from process result"""

        return Context._generate_response_cms(content, self.response_type, self.status_code, self.mime, self.cms, self.__headers)



import copy


class DictEx(dict):
    """Extended version of dict class for dot.notation access to attributes"""

    def __init__(self, *args, **kwargs):
        super().__init__()
        for arg in args:
            if isinstance(arg, dict):
                DictEx.fill_from_dic(self, arg)

    def has(self, key: str) -> bool:
        return key in self

    @classmethod
    def create(cls, data: dict):
        ret_val = DictEx()
        for k, v in data.items():
            if isinstance(v, dict):
                ret_val[k] = DictEx.create(v)
            elif isinstance(v, list):
                ret_val[k] = DictEx.file_from_list(v)
            else:
                ret_val[k] = v
        return ret_val

    @classmethod
    def fill_from_dic(cls, new, data: dict):
        for k, v in data.items():
            if isinstance(v, dict):
                new[k] = DictEx.create(v)
            elif isinstance(v, list):
                new[k] = DictEx.file_from_list(v)
            else:
                new[k] = v

    @classmethod
    def file_from_list(cls, data: list):
        ret_val_list = list()
        for item in data:
            if isinstance(item, dict):
                ret_val_list.append(DictEx.create(item))
            elif isinstance(item, list):
                ret_val_list.append(DictEx.file_from_list(item))
            else:
                ret_val_list.append(item)
        return ret_val_list

    __getattr__ = dict.get

    def __deepcopy__(self, memo=None):
        return DictEx(copy.deepcopy(dict(self), memo=memo))

    def __copy__(self):
        return DictEx(copy.copy(dict(self)))



from bclib.parser.answer.storage_data import StorageData


class EnrichedData:
    def __init__(self, part_number:"int", data_type:"str", storage_data:"StorageData|None", validators:"dict") -> None:
        self.part = part_number
        self.data_type = data_type
        self.storage_data = storage_data
        self.validators = validators


"""Base class for dispatching request"""
import asyncio
import inspect
from abc import ABC
import signal
import sys
import traceback
from typing import Callable, Any, Coroutine
from functools import wraps

from bclib.logger import ILogger, LoggerFactory
from bclib.cache import CacheFactory
from bclib.listener import RabbitBusListener
from bclib.predicate import Predicate
from bclib.context import ClientSourceContext, ClientSourceMemberContext, WebContext, Context, RESTfulContext, RabbitContext, SocketContext, ServerSourceContext, ServerSourceMemberContext, NamedPipeContext
from bclib.db_manager import DbManager
from bclib.utility import DictEx
from bclib.exception import HandlerNotFoundErr
from ..dispatcher.callback_info import CallbackInfo


class Dispatcher(ABC):
    """Base class for dispatching request"""

    def __init__(self, options: dict = None):
        self.options = DictEx(options)
        self.__look_up: 'dict[str, list[CallbackInfo]]' = dict()
        cache_options = self.options.cache if "cache" in self.options else None
        if sys.platform == 'win32':
            # By default Windows can use only 64 sockets in asyncio loop. This is a limitation of underlying select() API call.
            # Use Windows version of proactor event loop using IOCP
            loop = asyncio.ProactorEventLoop()
            asyncio.set_event_loop(loop)
        self.event_loop = asyncio.get_event_loop()
        self.cache_manager = CacheFactory.create(cache_options)
        self.db_manager = DbManager(self.options, self.event_loop)
        self.__logger: ILogger = LoggerFactory.create(self.options)
        self.log_error: bool = self.options.log_error if self.options.has(
            "log_error") else False
        self.log_request: bool = self.options.log_request if self.options.has(
            "log_request") else True
        self.__rabbit_dispatcher: 'list[RabbitBusListener]' = list()
        if "router" in self.options and "rabbit" in self.options.router:
            for setting in self.options.router.rabbit:
                self.__rabbit_dispatcher.append(
                    RabbitBusListener(setting, self))

    def socket_action(self, * predicates: (Predicate)):
        """Decorator for determine Socket action"""

        def _decorator(socket_action_handler: 'Callable[[SocketContext],bool]'):

            @wraps(socket_action_handler)
            async def non_async_wrapper(context: SocketContext):
                await self.event_loop.run_in_executor(None, socket_action_handler, context)
                return True

            @wraps(socket_action_handler)
            async def async_wrapper(context: SocketContext):
                await socket_action_handler(context)
                return True

            wrapper = async_wrapper if inspect.iscoroutinefunction(
                socket_action_handler) else non_async_wrapper

            self._get_context_lookup(SocketContext.__name__)\
                .append(CallbackInfo([*predicates], wrapper))
            return socket_action_handler
        return _decorator

    def restful_action(self, * predicates: (Predicate)):
        """Decorator for determine RESTful action"""

        def _decorator(restful_action_handler: 'Callable[[RESTfulContext], dict]'):

            @wraps(restful_action_handler)
            async def non_async_wrapper(context: RESTfulContext):
                action_result = await self.event_loop.run_in_executor(None, restful_action_handler, context)
                return None if action_result is None else context.generate_response(action_result)

            @wraps(restful_action_handler)
            async def async_wrapper(context: RESTfulContext):
                action_result = await restful_action_handler(context)
                return None if action_result is None else context.generate_response(action_result)

            wrapper = async_wrapper if inspect.iscoroutinefunction(
                restful_action_handler) else non_async_wrapper

            self._get_context_lookup(RESTfulContext.__name__)\
                .append(CallbackInfo([*predicates], wrapper))
            return restful_action_handler
        return _decorator

    def web_action(self, * predicates: (Predicate)):
        """Decorator for determine legacy web request action"""

        def _decorator(web_action_handler: 'Callable[[WebContext], str]'):

            @wraps(web_action_handler)
            async def non_async_wrapper(context: WebContext):
                action_result = await self.event_loop.run_in_executor(None, web_action_handler, context)
                return None if action_result is None else context.generate_response(action_result)

            @wraps(web_action_handler)
            async def async_wrapper(context: WebContext):
                action_result = await web_action_handler(context)
                return None if action_result is None else context.generate_response(action_result)

            wrapper = async_wrapper if inspect.iscoroutinefunction(
                web_action_handler) else non_async_wrapper

            self._get_context_lookup(WebContext.__name__)\
                .append(CallbackInfo([*predicates], wrapper))
            return web_action_handler
        return _decorator

    def client_source_action(self, *predicates: (Predicate)):
        """Decorator for determine source action"""

        def _decorator(client_source_action_handler: 'Callable[[ClientSourceContext], Any]'):
            @wraps(client_source_action_handler)
            async def non_async_wrapper(context: ClientSourceContext):
                data = await self.event_loop.run_in_executor(None, client_source_action_handler, context)
                result_set = list()
                if data is not None:
                    for member in context.command.member:
                        member_context = ClientSourceMemberContext(
                            context, data, member)
                        dispatch_result = await self.dispatch_async(member_context)
                        result = {
                            "options": {
                                "tableName": member_context.table_name,
                                "keyFieldName": member_context.key_field_name,
                                "statusFieldName": member_context.status_field_name,
                                "mergeType": member_context.merge_type.value,
                                "columnNames": member_context.column_names,
                            },
                            "data": dispatch_result
                        }
                        result_set.append(result)
                    ret_val = {
                        "setting": {
                            "keepalive": False,
                        },
                        "sources": result_set
                    }
                    return context.generate_response(ret_val)
                else:
                    return None

            @wraps(client_source_action_handler)
            async def async_wrapper(context: ClientSourceContext):
                data = await client_source_action_handler(context)
                result_set = list()
                if data is not None:
                    for member in context.command.member:
                        member_context = ClientSourceMemberContext(
                            context, data, member)
                        dispatch_result = await self.dispatch_async(member_context)
                        result = {
                            "options": {
                                "tableName": member_context.table_name,
                                "keyFieldName": member_context.key_field_name,
                                "statusFieldName": member_context.status_field_name,
                                "mergeType": member_context.merge_type.value,
                                "columnNames": member_context.column_names,
                            },
                            "data": dispatch_result
                        }
                        result_set.append(result)
                    ret_val = {
                        "setting": {
                            "keepalive": False,
                        },
                        "sources": result_set
                    }
                    return context.generate_response(ret_val)
                else:
                    return None

            wrapper = async_wrapper if inspect.iscoroutinefunction(
                client_source_action_handler) else non_async_wrapper

            self._get_context_lookup(ClientSourceContext.__name__)\
                .append(CallbackInfo([*predicates], wrapper))

            return client_source_action_handler
        return _decorator

    def client_source_member_action(self, *predicates: (Predicate)):
        """Decorator for determine source member action method"""

        def _decorator(client_source_member_handler: 'Callable[[ClientSourceMemberContext], Any]'):

            @wraps(client_source_member_handler)
            async def non_async_wrapper(context: WebContext):
                return await self.event_loop.run_in_executor(None, client_source_member_handler, context)

            @wraps(client_source_member_handler)
            async def async_wrapper(context: WebContext):
                return await client_source_member_handler(context)

            wrapper = async_wrapper if inspect.iscoroutinefunction(
                client_source_member_handler) else non_async_wrapper

            self._get_context_lookup(ClientSourceMemberContext.__name__)\
                .append(CallbackInfo([*predicates], wrapper))
            return client_source_member_handler
        return _decorator

    def server_source_action(self, *predicates: (Predicate)):
        """Decorator for determine source action"""

        def _decorator(server_source_action_handler: 'Callable[[ServerSourceContext], Any]'):
            @wraps(server_source_action_handler)
            async def non_async_wrapper(context: ServerSourceContext):
                data = await self.event_loop.run_in_executor(None, server_source_action_handler, context)
                result_set = list()
                if data is not None:
                    for member in context.command.member:
                        member_context = ServerSourceMemberContext(
                            context, data, member)
                        dispatch_result = await self.dispatch_async(member_context)
                        result = {
                            "options": {
                                "tableName": member_context.table_name,
                                "keyFieldName": member_context.key_field_name,
                                "statusFieldName": member_context.status_field_name,
                                "mergeType": member_context.merge_type.value,
                                "columnNames": member_context.column_names,
                            },
                            "data": dispatch_result
                        }
                        result_set.append(result)
                    ret_val = {
                        "setting": {
                            "keepalive": False,
                        },
                        "sources": result_set
                    }
                    return context.generate_response(ret_val)
                else:
                    return None

            @wraps(server_source_action_handler)
            async def async_wrapper(context: ServerSourceContext):
                data = await server_source_action_handler(context)
                result_set = list()
                if data is not None:
                    for member in context.command.member:
                        member_context = ServerSourceMemberContext(
                            context, data, member)
                        dispatch_result = await self.dispatch_async(member_context)
                        result = {
                            "options": {
                                "tableName": member_context.table_name,
                                "keyFieldName": member_context.key_field_name,
                                "statusFieldName": member_context.status_field_name,
                                "mergeType": member_context.merge_type.value,
                                "columnNames": member_context.column_names,
                            },
                            "data": dispatch_result
                        }
                        result_set.append(result)
                    ret_val = {
                        "setting": {
                            "keepalive": False,
                        },
                        "sources": result_set
                    }
                    return context.generate_response(ret_val)
                else:
                    return None

            wrapper = async_wrapper if inspect.iscoroutinefunction(
                server_source_action_handler) else non_async_wrapper

            self._get_context_lookup(ServerSourceContext.__name__)\
                .append(CallbackInfo([*predicates], wrapper))

            return server_source_action_handler
        return _decorator

    def server_source_member_action(self, *predicates: (Predicate)):
        """Decorator for determine server source member action method"""

        def _decorator(server_source_member_action_handler: 'Callable[[ServerSourceMemberContext], Any]'):

            @wraps(server_source_member_action_handler)
            async def non_async_wrapper(context: WebContext):
                return await self.event_loop.run_in_executor(None, server_source_member_action_handler, context)

            @wraps(server_source_member_action_handler)
            async def async_wrapper(context: WebContext):
                return await server_source_member_action_handler(context)

            wrapper = async_wrapper if inspect.iscoroutinefunction(
                server_source_member_action_handler) else non_async_wrapper

            self._get_context_lookup(ServerSourceMemberContext.__name__)\
                .append(CallbackInfo([*predicates], wrapper))
            return server_source_member_action_handler
        return _decorator

    def rabbit_action(self, * predicates: (Predicate)):
        """Decorator for determine rabbit-mq message request action"""

        def _decorator(rabbit_action_handler: 'Callable[[RabbitContext], bool]'):

            @wraps(rabbit_action_handler)
            async def non_async_wrapper(context: RabbitContext):
                return await self.event_loop.run_in_executor(None, rabbit_action_handler, context)

            @wraps(rabbit_action_handler)
            async def async_wrapper(context: RabbitContext):
                return await rabbit_action_handler(context)

            wrapper = async_wrapper if inspect.iscoroutinefunction(
                rabbit_action_handler) else non_async_wrapper

            self._get_context_lookup(RabbitContext.__name__)\
                .append(CallbackInfo([*predicates], wrapper))

            return rabbit_action_handler
        return _decorator

    def named_pipe_action(self, * predicates: (Predicate)):
        """Decorator for determine named pipe message request action"""

        def _decorator(named_pipe_action_handler: 'Callable[[RabbitContext], bool]'):

            @wraps(named_pipe_action_handler)
            async def non_async_wrapper(context: NamedPipeContext):
                return await self.event_loop.run_in_executor(None, named_pipe_action_handler, context)

            @wraps(named_pipe_action_handler)
            async def async_wrapper(context: NamedPipeContext):
                return await named_pipe_action_handler(context)

            wrapper = async_wrapper if inspect.iscoroutinefunction(
                named_pipe_action_handler) else non_async_wrapper

            self._get_context_lookup(NamedPipeContext.__name__)\
                .append(CallbackInfo([*predicates], wrapper))

            return named_pipe_action_handler
        return _decorator

    def _get_context_lookup(self, key: str) -> 'list[CallbackInfo]':
        """Get key related action list object"""

        ret_val: None
        if key in self.__look_up:
            ret_val = self.__look_up[key]
        else:
            ret_val = list()
            self.__look_up[key] = ret_val
        return ret_val

    async def dispatch_async(self, context: Context) -> Any:
        """Dispatch context and get result from related action method"""

        result: Any = None
        name = type(context).__name__
        try:
            items = self._get_context_lookup(name)
            for item in items:
                result = await item.try_execute_async(context)
                if result is not None:
                    break
            else:
                ex = HandlerNotFoundErr(name)
                if self.log_error:
                    print(str(ex))
                result = context.generate_error_response(ex)
        except Exception as ex:
            if self.log_error:
                traceback.print_exc()
            result = context.generate_error_response(ex)
        return result

    def dispatch_in_background(self, context: 'Context') -> asyncio.Future:
        """Dispatch context in background"""

        return self.event_loop.create_task(self.dispatch_async(context))

    def initialize_task(self):
        for dispatcher in self.__rabbit_dispatcher:
            dispatcher.initialize_task(self.event_loop)

    def listening(self, before_start: Coroutine=None, after_end: Coroutine=None):
        """Start listening to request for process"""
        for sig in (signal.SIGTERM, signal.SIGINT):
            signal.signal(sig, lambda sig, _: self.event_loop.stop())
        if before_start != None:
            self.event_loop.run_until_complete(self.event_loop.create_task(before_start()))
        self.initialize_task()
        self.event_loop.run_forever()
        tasks = asyncio.all_tasks(loop=self.event_loop)
        for task in tasks:
            task.cancel()
        group = asyncio.gather(*tasks, return_exceptions=True)
        self.event_loop.run_until_complete(group)
        if after_end != None:
            self.event_loop.run_until_complete(self.event_loop.create_task(after_end()))
        self.event_loop.close()

    async def log_async(self, **kwargs):
        """log params"""
        await self.__logger.log_async(**kwargs)

    def log_in_background(self, **kwargs) -> Coroutine:
        """log params in background precess"""
        return self.event_loop.create_task(
            self.__logger.log_async(**kwargs))



from bclib.utility import DictEx
from abc import ABC, abstractmethod
from typing import Callable

class BaseSignaler(ABC):
    def __init__(self, reset_cache_callback:"Callable", options:"DictEx"=None) -> None:
        super().__init__()
        self._callback = reset_cache_callback
        self._options = options if options is not None else DictEx({})



from abc import ABC
import time

class BaseCacheItem(ABC):
    def __init__(self, data:"any", life_time:"int") -> None:
        super().__init__()
        self.__data = data
        self.__life_time = life_time
        self.__created_time = time.time()

    def __is_expired(self):
        return time.time() > (self.__created_time + self.__life_time) if self.__life_time > 0 else False
    
    def _update_data(self, data):
        self.__data = data
        self.__created_time = time.time()
    
    def data(self):
        if self.__data is not None and self.__is_expired():
            self.__data = None
        return self.__data




from ..answer.enriched_data import EnrichedData

class QuestionData:
    def __init__(self, prpid:"int", ownerid:"int|None", typeid:"int|None", wordid:"int|None", enriched_data:"list[EnrichedData]") -> None:
        self.prpid = prpid
        self.ownerid = ownerid if ownerid is not None else 0
        self.typeid = typeid
        self.wordid = wordid
        self.enriched_data = enriched_data


from bclib.context.client_source_context import ClientSourceContext
from bclib.context.client_source_member_context import ClientSourceMemberContext
from bclib.context.context import Context
from bclib.context.restful_context import RESTfulContext
from bclib.context.web_context import WebContext
from bclib.context.request_context import RequestContext
from bclib.context.rabbit_context import RabbitContext
from bclib.context.socket_context import SocketContext
from bclib.context.merge_type import MergeType
from bclib.context.server_source_context import ServerSourceContext
from bclib.context.server_source_member_context import ServerSourceMemberContext
from bclib.context.source_context import SourceContext
from bclib.context.source_member_context import SourceMemberContext
from bclib.context.named_pipe_context import NamedPipeContext



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bclib.context import RequestContext


class HttpHeaders:
    ACCESS_CONTROL_ALLOW_CREDENTIALS = "Access-Control-Allow-Credentials"
    ACCESS_CONTROL_ALLOW_HEADERS = "Access-Control-Allow-Headers"
    ACCESS_CONTROL_ALLOW_METHODS = "Access-Control-Allow-Methods"
    ACCESS_CONTROL_ALLOW_ORIGIN = "Access-Control-Allow-Origin"
    ACCESS_CONTROL_EXPOSE_HEADERS = "Access-Control-Expose-Headers"
    ACCESS_CONTROL_MAX_AGE = "Access-Control-Max-Age"
    ACCEPT_RANGES = "Accept-Ranges"
    AGE = "Age"
    ALLOW = "Allow"
    ALTERNATE_PROTOCOL = "Alternate-Protocol"
    CACHE_CONTROL = "Cache-Control"
    CLIENT_DATE = "Client-Date"
    CLIENT_PEER = "Client-Peer"
    CLIENT_RESPONSE_NUM = "Client-Response-Num"
    CONNECTION = "Connection"
    CONTENT_DISPOSITION = "Content-Disposition"
    CONTENT_ENCODING = "Content-Encoding"
    CONTENT_LANGUAGE = "Content-Language"
    CONTENT_LENGTH = "Content-Length"
    CONTENT_LOCATION = "Content-Location"
    CONTENT_MD5 = "Content-MD5"
    CONTENT_RANGE = "Content-Range"
    CONTENT_SECURITY_POLICY = "Content-Security-Policy"
    X_CONTENT_SECURITY_POLICY = "X-Content-Security-Policy"
    X_WEBKIT_CSP = "X-WebKit-CSP"
    CONTENT_SECURITY_POLICY_REPORT_ONLY = "Content-Security-Policy-Report-Only"
    CONTENT_TYPE = "Content-Type"
    DATE = "Date"
    ETAG = "ETag"
    EXPIRES = "Expires"
    HTTP = "HTTP"
    KEEP_ALIVE = "Keep-Alive"
    LAST_MODIFIED = "Last-Modified"
    LINK = "Link"
    LOCATION = "Location"
    P3P = "P3P"
    PRAGMA = "Pragma"
    PROXY_AUTHENTICATE = "Proxy-Authenticate"
    PROXY_CONNECTION = "Proxy-Connection"
    REFRESH = "Refresh"
    RETRY_AFTER = "Retry-After"
    SERVER = "Server"
    SET_COOKIE = "Set-Cookie"
    STATUS = "Status"
    STRICT_TRANSPORT_SECURITY = "Strict-Transport-Security"
    TIMING_ALLOW_ORIGIN = "Timing-Allow-Origin"
    TRAILER = "Trailer"
    TRANSFER_ENCODING = "Transfer-Encoding"
    UPGRADE = "Upgrade"
    VARY = "Vary"
    VIA = "Via"
    WARNING = "Warning"
    WWW_AUTHENTICATE = "WWW-Authenticate"
    X_ASPNET_VERSION = "X-Aspnet-Version"
    X_CONTENT_TYPE_OPTIONS = "X-Content-Type-Options"
    X_FRAME_OPTIONS = "X-Frame-Options"
    X_PERMITTED_CROSS_DOMAIN_POLICIES = "X-Permitted-Cross-Domain-Policies"
    X_PINGBACK = "X-Pingback"
    X_POWERED_BY = "X-Powered-By"
    X_ROBOTS_TAG = "X-Robots-Tag"
    X_UA_COMPATIBLE = "X-UA-Compatible"
    X_XSS_PROTECTION = "X-XSS-Protection"

    @staticmethod
    def add_cors_headers(context: 'RequestContext'):
        context.add_header(HttpHeaders.ACCESS_CONTROL_ALLOW_ORIGIN, "*")
        context.add_header(HttpHeaders.ACCESS_CONTROL_ALLOW_HEADERS,
                           "Content-Type, Access-Control-Allow-Headers, Authorization, X-Requested-With")



from abc import ABC, abstractmethod
from bclib.utility import DictEx
from ..cache.cache_status import CacheStatus

class CacheManager(ABC):
    def __init__(self, options:"DictEx") -> None:
        super().__init__()
        self._options = options

    @abstractmethod
    def cache_decorator(self, key:"str"=None, life_time:"int"=0): ...

    @abstractmethod
    def get_cache(self, key:"str") -> "list|any|None": ...
    
    @abstractmethod
    def add_or_update(self, key:"str", data:"any", life_time:"int"=0) -> "CacheStatus": ...
    
    @abstractmethod
    def clean(self) -> "CacheStatus": ...

    @abstractmethod
    def reset(self, keys:"list[str]"=None) -> "CacheStatus": ...



from bclib.predicate.equal import Equal
from bclib.predicate.in_list import InList
from bclib.predicate.predicate import Predicate
from bclib.predicate.url import Url
from bclib.predicate.between import Between
from bclib.predicate.not_equal import NotEqual
from bclib.predicate.greater_than import GreaterThan
from bclib.predicate.less_than import LessThan
from bclib.predicate.less_than_equal import LessThanEqual
from bclib.predicate.greater_than_equal import GreaterThanEqual
from bclib.predicate.match import Match
from bclib.predicate.has_value import HasValue
from bclib.predicate.callback import Callback
from bclib.predicate.all import All
from bclib.predicate.any import Any



from ..cache.manager import CacheManager
from ..cache.cache_status import CacheStatus


class NoCacheManager(CacheManager):
    """"Implementing non caching. Only palace holder for None setting"""

    def cache_decorator(self, key: str = None, life_time: int = 0):
        def decorator(function):
            return function
        return decorator

    def get_cache(self, key: str) -> "list|None":
        return None

    def add_or_update(self, key: str, data: "any", life_time: int = 0) -> CacheStatus:
        return CacheStatus.ERROR

    def clean(self) -> None:
        pass

    def reset(self, keys: 'list[str]' = None) -> None:
        pass



import asyncio
from typing import Callable, Coroutine
from ..listener.message import Message
from bclib.utility import WindowsNamedPipeHelper


class WindowsNamedPipeListener:
    """"""

    def __init__(self, pipe_name: str, on_message_receive_call_back: 'Callable[[Message], Coroutine[Message]]'):
        self.pipe_name = pipe_name
        self.on_message_receive = on_message_receive_call_back
        self.__writer_pipe = None
        self.__reader_pipe = None
        self.__event_loop: asyncio.AbstractEventLoop = None

    async def __connect_writer_pipe_async(self):
        import win32pipe
        import pywintypes
        try:
            name = F"{self.pipe_name}/writer"
            self.__writer_pipe = win32pipe.CreateNamedPipe(name, win32pipe.PIPE_ACCESS_OUTBOUND,
                                                           win32pipe.PIPE_TYPE_MESSAGE | win32pipe.PIPE_READMODE_MESSAGE | win32pipe.PIPE_WAIT,
                                                           1, 65536, 65536, 0, None)
            print(
                f"Writer named pipe '{name}' is created. Waiting for reader client to connect...")
            await WindowsNamedPipeHelper.wait_for_client_connect_async(self.__writer_pipe, self.__event_loop)
            print(
                f"Reader client is connected to '{name}'...")
        except pywintypes.error as e:  # pylint: disable=maybe-no-member
            self.__writer_pipe = None
            if e.args[0] == 2:   # ERROR_FILE_NOT_FOUND
                print(f"No Named Pipe.  {repr(e)}")
            else:
                print(f"Named Pipe error code {e.args[0]}. {repr(e)}")
            raise
        except Exception as ex:
            self.__writer_pipe = None
            print(f"Error in create writer named pipe. {repr(ex)}")
            raise

    async def __process_message_async(self, message: 'Message') -> None:
        result = await self.on_message_receive(message)
        if result:
            await self.send_message_async(result)

    async def send_message_async(self, message: Message) -> bool:
        try_count = 0
        send = False
        while not send:
            try:
                if self.__writer_pipe is None:
                    await self.__connect_writer_pipe_async()
                WindowsNamedPipeHelper.write_to_named_pipe(
                    message, self.__writer_pipe)
                send = True
            except asyncio.CancelledError:
                break
            except Exception as ex:
                try_count = try_count+1
                self.__writer_pipe = None
                print(f"Error in send message {ex}")
                if try_count > 3:
                    break
                await asyncio.sleep(.5)
        return send

    def initialize_task(self, loop: asyncio.AbstractEventLoop):
        self.__event_loop = loop

        async def reader_loop_async():
            import win32pipe
            import win32file
            import pywintypes
            while True:
                try:
                    name = F"{self.pipe_name}/reader"
                    self.__reader_pipe = win32pipe.CreateNamedPipe(name, win32pipe.PIPE_ACCESS_INBOUND,
                                                                   win32pipe.PIPE_TYPE_MESSAGE | win32pipe.PIPE_READMODE_MESSAGE | win32pipe.PIPE_WAIT,
                                                                   1, 65536, 65536, 0, None)
                    print(
                        f"Reader named pipe '{name}' is created. Waiting for writer client to connect...")
                    await WindowsNamedPipeHelper.wait_for_client_connect_async(self.__reader_pipe, self.__event_loop)
                    print(
                        f"Writer client connect to '{name}'...")
                    while True:
                        message = await WindowsNamedPipeHelper.read_from_named_pipe_async(
                            self.__reader_pipe, self.__event_loop)
                        if message:
                            self.__event_loop.create_task(
                                self.__process_message_async(message))
                except asyncio.CancelledError:
                    print('Edge named pipe server stopped.!')
                    break
                except pywintypes.error as e:  # pylint: disable=maybe-no-member
                    if e.args[0] == 2:   # ERROR_FILE_NOT_FOUND
                        print(f"No reader named pipe.  {repr(e)}")
                    elif e.args[0] == 109:   # ERROR_BROKEN_PIPE
                        print(f"Reader named pipe is broken. {repr(e)}")
                    else:
                        print(
                            f"Reader named pipe error code {e.args[0]}. {repr(e)}")
                finally:
                    # Disconnect the named pipe
                    win32pipe.DisconnectNamedPipe(self.__reader_pipe)
                    # CLose the named pipe
                    win32file.CloseHandle(self.__reader_pipe)
        self.__event_loop.create_task(reader_loop_async())
        self.__event_loop.create_task(self.__connect_writer_pipe_async())



import asyncio
from sys import platform
from ..named_pipe.inamed_pipe_connection import INamedPipeConnection


class NamedPipeConnection:
    __dic: 'dict[str,INamedPipeConnection]' = dict()

    @staticmethod
    def get_connection(pipe_name: str, loop: asyncio.AbstractEventLoop) -> INamedPipeConnection:
        ret_val = None
        if pipe_name in NamedPipeConnection.__dic:
            ret_val = NamedPipeConnection.__dic[pipe_name]
        else:
            # https://docs.python.org/3/library/sys.html#sys.platform
            if platform == "linux" or platform == "linux2":
                # linux
                from ..named_pipe.linux_named_pipe_connection import LinuxNamedPipeConnection
                ret_val = NamedPipeConnection.__dic[pipe_name] = LinuxNamedPipeConnection(
                    pipe_name, loop)
            elif platform == "darwin":
                # OS X
                raise Exception(
                    "named pipe connection not implemented in OS X")
            elif platform == "win32":
                from ..named_pipe.windows_named_pipe_connection import WindowsNamedPipeConnection
                ret_val = NamedPipeConnection.__dic[pipe_name] = WindowsNamedPipeConnection(
                    pipe_name, loop)
            NamedPipeConnection.__dic[pipe_name] = ret_val
        return ret_val



import asyncio
from io import BufferedWriter
import os
import sys
from typing import Callable, Coroutine
from ..listener.message import Message
from bclib.utility import LinuxNamedPipeHelper


class LinuxNamedPipeListener:
    """"""

    def __init__(self, pipe_name: str, on_message_receive_call_back: 'Callable[[Message], Coroutine[Message]]'):
        self.pipe_name = pipe_name
        self.__writer_pipe_name = F"{self.pipe_name}-writer"
        self.on_message_receive = on_message_receive_call_back
        self.__writer_pipe: BufferedWriter = None
        self.__event_loop: asyncio.AbstractEventLoop = None

    def try_unlink(self, pipe_name: str):
        try:
            os.unlink(pipe_name)
            print(f"unlink named pipe '{pipe_name}'...")
        except:
            pass

    async def __connect_writer_pipe_async(self):
        if self.__writer_pipe:
            self.try_unlink(self.__writer_pipe_name)

        try:
            os.mkfifo(self.__writer_pipe_name)
        except Exception as ex:
            print('error in run os.mkfifo():', ex)

        try:
            self.__writer_pipe = open(self.__writer_pipe_name, "wb")
            print(
                f"Writer named pipe '{self.__writer_pipe_name}' is created...")
        except Exception as ex:
            print(f"Error in create writer named pipe. {repr(ex)}")
            self.try_unlink(self.__writer_pipe_name)
            raise

    async def __process_message_async(self, message: 'Message') -> None:
        result = await self.on_message_receive(message)
        if result:
            await self.send_message_async(result)

    async def send_message_async(self, message: Message) -> bool:
        try_count = 0
        send = False
        while not send:
            try:
                if self.__writer_pipe is None:
                    await self.__connect_writer_pipe_async()
                LinuxNamedPipeHelper.write_to_named_pipe(
                    message, self.__writer_pipe)
                send = True
            except asyncio.CancelledError:
                break
            except Exception as ex:
                try_count = try_count+1
                if self.__writer_pipe:
                    self.try_unlink(self.__writer_pipe_name)
                    try:
                        self.__writer_pipe.close()
                    except:
                        pass
                    self.__writer_pipe = None
                print(f"Error in send message {ex}")
                if try_count > 3:
                    break
                await asyncio.sleep(.5)
        return send

    def initialize_task(self, loop: asyncio.AbstractEventLoop):
        self.__event_loop = loop

        async def reader_loop_async():
            reader_pipe_name = F"{self.pipe_name}-reader"
            while True:
                try:
                    os.mkfifo(reader_pipe_name)
                except Exception as ex:
                    print(
                        f'Error in call os.mkfifo() for ${reader_pipe_name}', ex)

                with open(reader_pipe_name, "rb") as reader_pipe:
                    print(
                        f"Reader named pipe '{reader_pipe_name}' is created. try to read from it...")
                    try:
                        while True:
                            message = await LinuxNamedPipeHelper.read_from_named_pipe_async(reader_pipe, self.__event_loop)
                            if message:
                                self.__event_loop.create_task(
                                    self.__process_message_async(message))
                    except asyncio.CancelledError:
                        print('Edge named pipe server stopped.!')
                        break
                    except Exception as ex:
                        print('Error cause edge named pipe server restart!', ex)
                        await asyncio.sleep(1)
                    finally:
                        try:
                            if reader_pipe:
                                os.unlink(reader_pipe_name)
                                print("unlink...")
                        except Exception as ex:
                            print("Error in unlink named pipe...", ex)

        self.__event_loop.create_task(reader_loop_async())
        self.__event_loop.create_task(self.__connect_writer_pipe_async())



import asyncio
from sys import platform
from ..dispatcher.routing_dispatcher import RoutingDispatcher
from bclib.listener import Message


class NamedPipeDispatcher(RoutingDispatcher):
    def __init__(self, options: dict):
        super().__init__(options)
        self.__lock = asyncio.Lock()
        # https://docs.python.org/3/library/sys.html#sys.platform
        if platform == "linux" or platform == "linux2":
            # linux
            from bclib.listener import LinuxNamedPipeListener
            self.__listener = LinuxNamedPipeListener(
                self.options.named_pipe,
                self._on_message_receive_async)
        elif platform == "win32":
            from bclib.listener import WindowsNamedPipeListener
            self.__listener = WindowsNamedPipeListener(
                self.options.named_pipe,
                self._on_message_receive_async)
        elif platform == "darwin":
            # OS X
            raise Exception("named pipe not implemented in OS X")

    async def send_message_async(self, message: Message) -> bool:
        """Send message to endpoint"""

        async with self.__lock:
            return await self.__listener.send_message_async(message)

    def initialize_task(self):
        super().initialize_task()
        self.__listener.initialize_task(self.event_loop)



import asyncio
from typing import Callable, Coroutine
from ..listener.message import Message
from ..listener.endpoint import Endpoint


class SocketListener:
    def __init__(self, receiver: Endpoint, sender: Endpoint, on_message_receive_call_back: 'Callable[[Message], Coroutine[Message]]'):
        self.__receiver_endpoint = receiver
        self.__sender_endpoint = sender
        self.on_message_receive = on_message_receive_call_back
        self.__sender_stream_writer: asyncio.StreamWriter = None
        self.__receiver_server: asyncio.AbstractServer = None
        self.__sender_server: asyncio.AbstractServer = None

    async def __process_message_async(self, message: 'Message') -> None:
        result = await self.on_message_receive(message)
        if result:
            await self.send_message_async(result)

    async def send_message_async(self, message: Message) -> bool:
        try:
            await message.write_to_stream_async(self.__sender_stream_writer)
        except Exception as ex:
            print(f"Error in send message {ex}")
            return False

    async def on_sender_client_connect(self, _: asyncio.StreamReader, writer: asyncio.StreamWriter):
        peer_name = writer.get_extra_info('peername')
        print(f'Reader from {peer_name}, connect to sender!')
        if self.__sender_stream_writer and not self.__sender_stream_writer.is_closing():
            if self.__sender_stream_writer.can_write_eof():
                self.__sender_stream_writer.write_eof()
                await self.__sender_stream_writer.drain()
            self.__sender_stream_writer.close()
            await self.__sender_stream_writer.wait_closed()
        self.__sender_stream_writer = writer
        cause = "closed!"
        try:
            while True:
                await asyncio.sleep(1)
        except asyncio.CancelledError:
            cause = 'closed by sender!'
        except (ConnectionResetError, asyncio.IncompleteReadError):
            cause = 'disconnected!'
        finally:
            print(
                f'Reader from {peer_name}, {cause}')
            if writer and not writer.is_closing():
                if writer.can_write_eof():
                    writer.write_eof()
                    await writer.drain()
                writer.close()
                await writer.wait_closed()

    async def on_receiver_client_connect(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        peer_name = writer.get_extra_info('peername')
        print(f'Writer from {peer_name}, connect to receiver!')
        loop = asyncio.get_running_loop()
        cause = "closed!"
        try:
            while True:
                message = await Message.read_from_stream_async(reader)
                if message:
                    loop.create_task(self.__process_message_async(message))
        except asyncio.CancelledError:
            cause = 'closed by receiver!'
        except (ConnectionResetError, asyncio.IncompleteReadError):
            cause = 'disconnected!'
        finally:
            print(
                f'Writer from {peer_name}, {cause}')
            if writer and not writer.is_closing():
                if writer.can_write_eof():
                    writer.write_eof()
                    await writer.drain()
                writer.close()
                await writer.wait_closed()

    def initialize_task(self, loop: asyncio.AbstractEventLoop):
        async def start_servers_async():
            loop = asyncio.get_running_loop()
            try:
                self.__sender_server = await asyncio.start_server(
                    self.on_sender_client_connect,
                    host=self.__sender_endpoint.url,
                    port=self.__sender_endpoint.port)
                print(
                    f'Sender server up in {self.__sender_endpoint.url}:{self.__sender_endpoint.port} and wait for reader connection...')
                self.__receiver_server = await asyncio.start_server(
                    self.on_receiver_client_connect,
                    host=self.__receiver_endpoint.url,
                    port=self.__receiver_endpoint.port)
                print(
                    f'Receiver server up in {self.__receiver_endpoint.url}:{self.__receiver_endpoint.port} and wait for writer connection...')
            except Exception as ex:
                try:
                    if self.__sender_server:
                        self.__sender_server.close()
                        await self.__sender_server.wait_closed()
                except:
                    pass
                try:
                    if self.__receiver_server:
                        self.__receiver_server.close()
                        await self.__receiver_server.wait_closed()
                except:
                    pass
                print(f"server not start. {repr(ex)}")
                return

            async def sender_loop_async():
                try:
                    async with self.__sender_server:
                        await self.__sender_server.serve_forever()
                except asyncio.CancelledError:
                    self.__sender_server.close()
                    await self.__sender_server.wait_closed()
                    print("Sender server shutdown...")

            async def receiver_loop_async():
                try:
                    async with self.__receiver_server:
                        await self.__receiver_server.serve_forever()
                except asyncio.CancelledError:
                    self.__receiver_server.close()
                    await self.__receiver_server.wait_closed()
                    print("Receiver server shutdown...")
            loop.create_task(sender_loop_async())
            loop.create_task(receiver_loop_async())
        loop.create_task(start_servers_async())



from ..cache_item.base_cache_item import BaseCacheItem
from typing import Callable

class FunctionCacheItem(BaseCacheItem):
    def __init__(self, data: "any", life_time:"int", function:"Callable") -> None:
        super().__init__(data, life_time)
        self.__function = function

    def get_data(self, *args, **kwargs) -> "any":
        data = self.data()
        if data is None:
            data = self.__function(*args, **kwargs)
            self._update_data(data)
        return data


class Endpoint:
    """Represent ip and port number for endpoint"""

    def __init__(self, url: str):
        parts = url.split(':')
        self.__url = parts[0]
        self.__port = int(parts[1]) if len(parts) >= 2 else 80

    @property
    def url(self) -> str:
        return self.__url

    @property
    def port(self) -> int:
        return self.__port



from typing import Any
from bclib.context import Context
from ..predicate.predicate import Predicate


class InList(Predicate):
    """Create list cheking predicate"""

    def __init__(self, expression: str, *items: Any) -> None:
        super().__init__(expression)
        self.__items = [*items]

    async def check_async(self, context: Context) -> bool:
        try:
            value = eval(self.exprossion, {}, {"context": context})
            return value in self.__items
        except:  # pylint: disable=bare-except
            return False



from typing import TYPE_CHECKING
from . import Context
if TYPE_CHECKING:
    from .. import dispatcher
from bclib.utility import DictEx


class NamedPipeContext(Context):
    """Context for named pipe request"""

    def __init__(self, named_pipe_message: 'dict', raw_message: str, dispatcher: 'dispatcher.IDispatcher'):
        super().__init__(dispatcher)
        self.message: DictEx = DictEx(named_pipe_message)
        self.raw_message = raw_message



import asyncio

from ..listener import Endpoint, ReceiveMessage
from .routing_dispatcher import RoutingDispatcher


class EndpointDispatcher(RoutingDispatcher):
    def __init__(self, options: dict):
        super().__init__(options)
        self.__endpoint = Endpoint(self.options.endpoint)

    def initialize_task(self):
        super().initialize_task()

        async def on_connection_open(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
            try:
                msg = await ReceiveMessage.read_from_stream_async(reader, writer)
                result = await self._on_message_receive_async(msg)
                await result.write_to_stream_async(writer)
            except:
                pass
            try:
                if writer.can_write_eof():
                    writer.write_eof()
                await writer.drain()
                writer.close()
            except:
                pass

        async def start_servers_async():
            server = await asyncio.start_server(on_connection_open, self.__endpoint.url, self.__endpoint.port,)
            print(
                f'Server up in {self.__endpoint.url}:{self.__endpoint.port} and wait for client connection...')
            try:
                while True:
                    await asyncio.sleep(1)
            except asyncio.CancelledError:
                pass
            print("Server shutdown...")
            server.close()

        self.event_loop.create_task(start_servers_async())



from bclib.utility.dict_ex import DictEx
from bclib.utility.http_status_codes import HttpStatusCodes
from bclib.utility.http_mime_types import HttpMimeTypes
from bclib.utility.response_types import ResponseTypes
from bclib.utility.http_headers import HttpHeaders
from bclib.utility.windows_named_pipe_helper import WindowsNamedPipeHelper
from bclib.utility.linux_named_pipe_helper import LinuxNamedPipeHelper



from ..predicate.predicate import Predicate
from bclib.context import Context


class All (Predicate):
    """Create all predicate list cheking predicate"""

    def __init__(self, *predicate: 'Predicate') -> None:
        super().__init__(None)
        self.__predicate_list = predicate

    async def check_async(self, context: Context) -> bool:
        is_ok = True
        try:
            for predicate in self.__predicate_list:
                if not await predicate.check_async(context):
                    is_ok = False
                    break
        except:  # pylint: disable=bare-except
            is_ok = False
        return is_ok



import asyncio
from abc import ABC, abstractmethod
from bclib.utility import DictEx
import pika
from pika.exceptions import AMQPConnectionError


class RabbitListener(ABC):
    def __init__(self, connection_options: DictEx) -> None:
        try:
            self.__param = pika.URLParameters(connection_options.url)
            self._host: str = self.__param.host
            self._queue_name: str = connection_options.queue
            self.__passive = connection_options.passive if connection_options.has(
                "passive") else False
            self.__durable = connection_options.durable if connection_options.has(
                "durable") else False
            self.__exclusive = connection_options.exclusive if connection_options.has(
                "exclusive") else False
            self.__auto_delete = connection_options.auto_delete if connection_options.has(
                "auto_delete") else False
            self.__retry_delay: int = int(
                connection_options.retry_delay) if connection_options.has("retry_delay") else 60
            self.__try_to_apply_connection()
        except Exception as ex:
            print(f"Error in config rabbit-mq ({ex})")
            raise ex

    def __try_to_apply_connection(self):
        print("Rabbit connection attemp...")
        self.__connection = pika.BlockingConnection(self.__param)
        self.__channel = self.__connection.channel()
        self.__channel.queue_declare(
            queue=self._queue_name,
            passive=self.__passive,
            durable=self.__durable,
            exclusive=self.__exclusive,
            auto_delete=self.__auto_delete
        )
        self.__channel.basic_consume(
            queue=self._queue_name, on_message_callback=lambda channel, method, properties, body: self.on_rabbit_message_received(body), auto_ack=True)

    @abstractmethod
    def on_rabbit_message_received(self, body):
        pass

    def initialize_task(self, loop: asyncio.AbstractEventLoop) -> asyncio.Future:
        loop.create_task(self.__consuming_task())

    async def __consuming_task(self):
        while True:
            try:
                print(
                    f'Rabbit listener waiting for messages from "{self._host}:{self._queue_name}."')
                loop = asyncio.get_running_loop()
                await loop.run_in_executor(None, self.__channel.start_consuming)
                break
            except asyncio.CancelledError:
                self.__channel.stop_consuming()
                self.__channel.cancel()
                try:
                    self.__channel.close()
                except:
                    pass
                try:
                    self.__connection.close()
                except:
                    pass
                break
            except AMQPConnectionError as ex:
                print("Rabbit listener disconnected!")
                await asyncio.sleep(self.__retry_delay)



class ShortCircuitErr(Exception):
    def __init__(self, status_code: 'str', error_code: 'str' = -1, message: 'str' = None, data: 'dict' = None):
        super().__init__(message if message else '')
        self.data = data
        self.status_code = status_code
        self.error_code = error_code



from bclib.utility.http_status_codes import HttpStatusCodes
from .short_circuit_err import ShortCircuitErr


class ForbiddenErr(ShortCircuitErr):
    def __init__(self, message: str = None, data: 'dict' = None):
        super().__init__(HttpStatusCodes.FORBIDDEN, 'http-403', message, data)



import json
from typing import Any, Callable

from bclib.parser.answer.enriched_data import EnrichedData
from bclib.parser.answer.storage_data import StorageData
from bclib.parser.answer.question_data import QuestionData
from bclib import parser
from bclib.db_manager import RESTfulConnection
from bclib.parser.answer.validators import Validator
from bclib.utility import DictEx
from ..answer.user_action_types import UserActionTypes
from ..answer.user_action import UserAction

class Answer:
    """BasisJsonParser is a tool to parse basis_core components json objects. This tool is developed based on
             basis_core key and values."""

    def __init__(self, data: 'str|Any', api_url: 'str' = None, check_validation:"bool"= False):
        self.json = json.loads(data) if isinstance(data, str) else data
        self.__answer_list: 'list[UserAction]' = None
        self.__api_connection = RESTfulConnection(
            api_url) if api_url else None
        self.check_validation = check_validation 

    async def __fill_answer_list_async(self):
        self.__answer_list = list()
        internal_prp_value_index = 1
        for data in self.json['properties']:
            multi = data['multi'] if 'multi' in data else None
            for action_type in UserActionTypes:
                if action_type.value in list(data.keys()):
                    prp_id = data['propId'] if action_type != UserActionTypes.ANSWERS else data["prpId"]
                    for actions in data[action_type.value]:
                        prp_value_id = actions['id'] if 'id' in actions.keys() else None
                        if 'parts' in actions.keys():
                            for parts in actions['parts']:
                                internal_prp_value_id = internal_prp_value_index
                                part_number = parts['part'] if "part" in parts.keys() else None
                                for values in parts['values']:
                                    value_id = values['id'] if "id" in values.keys() else None
                                    value = values['value']
                                    answer = parser.ParseAnswer(
                                        values["answer"]) if 'answer' in values.keys() else None
                                    self.__answer_list.append(
                                        UserAction(
                                            prp_id, action_type, prp_value_id, internal_prp_value_id, value_id, value, None, multi, part_number, answer
                                        )
                                    )
                        else:
                            internal_prp_value_id = internal_prp_value_index
                            self.__answer_list.append(
                                UserAction(
                                    prp_id, action_type,  prp_value_id, internal_prp_value_id, None, None, None, multi, None, None
                                )
                            )
                        internal_prp_value_index += 1

        await self.__enrich_data_async()

    async def __enrich_data_async(self):
        questions_info: "list[QuestionData]" = list()
        if self.__api_connection:
            questions_data = DictEx(await self.__api_connection.get_async())
            for data in questions_data.sources:
                for question in data.data:
                    for parts in question.questions:
                        enriched_data_list: "list[EnrichedData]" = [
                            self.__enrich_data(validations)
                            for validations in parts.parts
                        ]
                        questions_info.append(
                            QuestionData(parts.prpId, parts.OwnerID, parts.TypeID, parts.wordId, enriched_data_list)
                        )
        if len(questions_info) > 0:
            for values in self.__answer_list:
                for data in questions_info:
                    if int(data.prpid) == int(values.prp_id):
                        values.ownerid = data.ownerid
                        values.typeid = data.typeid
                        values.wordid = data.wordid
                        for enriched_data in data.enriched_data:
                            if (enriched_data.part == values.part) or (values.part is None):
                                values.datatype = enriched_data.data_type
                                storage_data = enriched_data.storage_data
                                if storage_data:
                                    values.database = storage_data.data_base
                                    values.table = storage_data.table
                                    values.field = storage_data.field
                                if self.check_validation and values.action != UserActionTypes.DELETED:
                                    status, message = Validator.check_validators(enriched_data.validators, values.value)
                                    values.validation_status = status
                                    values.validation_message = message
                
    def __enrich_data(self, validations:DictEx) -> 'EnrichedData':
        part_id = validations.part
        data_type = self.__set_data_type(validations)
        val_val = validations.validations
        storage_data = self.__set_storage_data(val_val) if isinstance(val_val, dict) else None
        validators = val_val if self.check_validation and isinstance(validations.validations, dict) else {}
        return EnrichedData(part_id, data_type, storage_data, validators)

    def __set_data_type(self, validations:DictEx) -> 'str':
        has_link = True if validations.link else False
        val_val = validations.validations
        data_type = val_val["dataType"] if isinstance(val_val, dict) and "dataType" in val_val else None
        
        return self.__data_type_checker(validations.viewType, data_type, has_link)

    def __data_type_checker(self, view_type: str, datatype: str = None, has_link: bool = None):
        if view_type in ["select", "checklist", "radio"]:
            if has_link == True:
                result = "urlvalue"
            else:
                result = "fixvalue"
        elif view_type == "textarea":
            result = "ntextvalue"
        elif view_type == "text" and datatype in ["text", "None", None]:
            result = "textvalue"
        elif view_type == "text" and datatype == "int":
            result = "numvalue"
        elif view_type == "text" and datatype == "float":
            result = "floatvalue"
        elif view_type == "autocomplete":
            result = "urlvalue"
        elif view_type == "upload":
            result = "files"
        elif view_type == "color":
            result = "textvalue"
        elif view_type == "reference":
            result = "reference"
        else:
            result = "None"
        return result
    
    def __set_storage_data(self, val_val:"dict") -> "StorageData":
        database = val_val["database"] if "database" in val_val else None
        table = val_val["table"] if "table" in val_val else None
        field = val_val["field"] if "field" in val_val else None
        return StorageData(database, table, field)
                
    async def __get_action_async(self, prp_id_list: 'list[int]', action_list: 'list[UserActionTypes]', part_list: 'list[int]', is_file: "bool" = None, predicate: 'Callable[[UserAction],bool]' = None) -> 'list[UserAction]':
        ret_val: 'list[UserAction]' = None
        if self.__answer_list is None:
            await self.__fill_answer_list_async()
        if predicate:
            if is_file == None:
                ret_val = [x for x in self.__answer_list if predicate(x)]
            else:
                ret_val = [x for x in self.__answer_list if predicate(
                    x) and x.is_file_content() == is_file]
        else:
            ret_val = [x for x in self.__answer_list if
                       (prp_id_list is None or x.prp_id in prp_id_list) and
                       (action_list is None or x.action in action_list) and
                       (part_list is None or x.part in part_list) and
                       (is_file is None or x.is_file_content() == is_file)
                       ]

        return ret_val if ret_val else list()

    async def get_actions_async(self, prp_id: 'int|list[int]' = None, action: 'UserActionTypes|list[UserActionTypes]' = None,
                                part: int = None, is_file: "bool" = None, predicate: 'Callable[[UserAction],bool]' = None) -> 'list[UserAction]':
        """
        inputs:
        prpid: None, int or list[int]
        action: None, int or list[int]
        part: None or int
        Samples:
        (prpid=None, action='edited')
        (prpid=[12345, 1000], action=None)
        """

        action_list = [action] if isinstance(
            action, UserActionTypes) else action
        prp_id_list = [prp_id] if isinstance(prp_id, int) else prp_id
        part_list = [part] if isinstance(part, int) else part
        return await self.__get_action_async(prp_id_list, action_list, part_list, is_file, predicate)

    async def get_added_actions_async(self, prp_id: 'int|list[int]' = None, predicate: 'Callable[[UserAction],bool]' = None) -> 'list[list[list[UserAction]]]':
        return await self.__get_specify_actions(UserActionTypes.ADDED, prp_id, predicate)

    async def get_edited_actions_async(self, prp_id: 'int|list[int]' = None, predicate: 'Callable[[UserAction],bool]' = None) -> 'list[list[list[UserAction]]]':
        return await self.__get_specify_actions(UserActionTypes.EDITED, prp_id, predicate)

    async def __get_specify_actions(self, action_type: UserActionTypes, prp_id: 'int|list[int]' = None, predicate: 'Callable[[UserAction],bool]' = None) -> 'list[list[list[UserAction]]]':
        ret_val: "dict[int,list[list[UserAction]]]" = {}
        action_objects = await self.get_actions_async(prp_id=prp_id, action=action_type, predicate=predicate)
        unique_internal_values_id = set(
            [obj.internal_prp_value_id for obj in action_objects])
        same_internal_value_id_objects = [[obj for obj in action_objects if obj.internal_prp_value_id ==
                                           internal_values_id] for internal_values_id in unique_internal_values_id]
        unique_prp_id = set(
            [obj[0].prp_id for obj in same_internal_value_id_objects])
        for _prp_id in unique_prp_id:
            for obj in same_internal_value_id_objects:
                if obj[0].prp_id == _prp_id:
                    if obj[0].prp_id in ret_val:
                        ret_val[obj[0].prp_id].append(obj)
                    else:
                        ret_val[obj[0].prp_id] = [obj]
        return list(ret_val.values())

    async def is_valid_answer_async(self):
        if self.__answer_list is None:
            await self.__fill_answer_list_async()
        is_valid = True
        for user_action in self.__answer_list:
            if user_action.validation_status == False:
                is_valid = False
                break
        return is_valid



from typing import Any


class LogSchema:
    def __init__(self, schema: 'dict[str,Any]') -> None:
        self.schema_name = schema["schemaName"]
        self.schema_version = schema["schemaVersion"]
        self.lid = schema["lid"]
        self.schemaId = schema["schemaId"]
        self.paramUrl = schema["paramUrl"]
        self.properties: 'dict[str,(int,str)]' = dict([
            (x["title"], (x["prpId"], x["TypeID"] if "TypeID" in x else 0, x["source"] if "source" in x else None)) for x in schema["questions"]])

    def get_answer(self, params: 'dict[str,Any]'):
        properties = list()
        for key, (id, typeid, source) in self.properties.items():
            value = None
            try:
                if source:
                    value = eval(source, {}, params)
                else:
                    value = params[key] if key in params else None
            except:
                pass

            if value is not None:
                properties.append({
                    "prpId": id,
                    "TypeID": typeid,
                    "answers": [
                        {
                            "parts": [
                                {
                                    "part": 1,
                                    "values": [
                                        {
                                            "value": value
                                        }
                                    ]
                                }
                            ]
                        }
                    ]
                })
        return {
            "schemaName": self.schema_name,
            "paramUrl": self.paramUrl,
            "schemaVersion": self.schema_version,
            "lid": self.lid,
            "schemaId": self.schemaId,
            "properties": properties
        }



import json
from typing import TYPE_CHECKING
from .context import Context
from bclib.listener.receive_message import ReceiveMessage
from bclib.utility import DictEx, HttpStatusCodes, HttpMimeTypes, HttpStatusCodes, ResponseTypes

if TYPE_CHECKING:
    from .. import dispatcher
    from .. import listener


class SocketContext(Context):
    """Base class for dispatching web socket base request context"""

    def __init__(self, cms_object: dict,  dispatcher: 'dispatcher.IDispatcher', message_object: 'listener.ReceiveMessage', body: dict) -> None:
        super().__init__(dispatcher)
        self.cms = DictEx(cms_object) if cms_object else None
        self.url: str = self.cms.request.url if cms_object else None
        self.body = DictEx(body) if body else None
        self.message = message_object
        self.is_adhoc = False

    async def send_object_async(self, obj) -> None:
        await self.send_content_async(json.dumps(obj))

    async def send_html_async(self, html) -> None:
        await self.send_content_async(html, mime=HttpMimeTypes.HTML)

    async def send_code_async(self, code) -> None:
        await self.send_content_async(code, response_type=ResponseTypes.RENDERABLE, mime=HttpMimeTypes.HTML)

    async def send_content_async(self,
                                 content: 'Any',
                                 response_type: 'str' = ResponseTypes.RENDERED,
                                 status_code: 'str' = HttpStatusCodes.OK,
                                 mime: 'str' = HttpMimeTypes.JSON,
                                 template: 'DictEx' = None,
                                 headers: 'dict' = None) -> None:
        cms = Context._generate_response_cms(
            content, response_type, status_code, mime, template, headers)
        message = ReceiveMessage.create_from_object(
            self.message.session_id, cms)
        await message.write_to_stream_async(self.message.writer)

    async def read_message_async(self) -> 'listener.ReceiveMessage':
        return await self.message.read_next_message_async()

    async def send_close_async(self) -> None:
        message = ReceiveMessage.create_disconnect(self.message.session_id)
        await message.write_to_stream_async(self.message.writer)



from bclib.context import Context
from ..predicate.predicate import Predicate


class Between(Predicate):
    """Create between cheking predicate"""

    def __init__(self, expression: str, min_value: int, max_value: int):
        super().__init__(expression)
        self.__min_value = min_value
        self.__max_value = max_value

    async def check_async(self, context: Context) -> bool:
        try:
            value = eval(self.exprossion, {}, {"context": context})
            return self.__min_value < int(value) < self.__max_value
        except:  # pylint: disable=bare-except
            return False



from ..exception.not_found_err import NotFoundErr


class HandlerNotFoundErr(NotFoundErr):
    def __init__(self, context_type: str = None):
        super().__init__(f"Suitable handler not found for {context_type}!")



import setuptools
from bclib import edge

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bclib",
    version=edge.__version__,
    author="Manzoomeh Negaran",
    author_email="info@manzoomeh.ir",
    description="Python base gateway for communicate with basiscore webserver",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Manzoomeh/BasisCore.Server.Edge/wiki",
    project_urls={
        "Bug Tracker": "https://github.com/Manzoomeh/BasisCore.Server.Edge/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    # install_requires=[
    #     'pika',
    #     'requests',
    #     'pymongo',
    #     'pyodbc'
    # ],
    #package_dir={"": "basiscore"},
    packages=setuptools.find_packages(exclude=["test", "app-env", ".vscode"]),
    python_requires=">=3.7",
    setup_requires=['wheel']
)



class FileUserAction:

    def __init__(self, prp_id: "int", prp_value_id: "int", value_id: "int", name: "str", type: "str", size: "int", content: "str") -> None:
        self.prp_id = prp_id
        self.prp_value_id = prp_value_id
        self.value_id = value_id
        self.name = name
        self.type = type
        self.size = size
        self.content = content

    def as_tuple(self) -> tuple:
        return (self.prp_id, self.prp_value_id, self.value_id, self.name, self.type, self.size, self.content)

    def as_dict(self) -> dict:
        return {
            "prp_id": self.prp_id,
            "prp_value_id": self.prp_value_id,
            "value_id": self.value_id,
            "name": self.name,
            "type": self.type,
            "size": self.size,
            "content": self.content
        }

    def __str__(self) -> str:
        return str(self.as_tuple())

    def __repr__(self) -> str:
        return str(self.as_dict())



from bclib.cache.cache_item.base_cache_item import BaseCacheItem
from ..value_item.base_value_item import BaseValueItem
from ..cache_item.function_cache_item import FunctionCacheItem

class ArrayValueItem(BaseValueItem):

    def _apply_item(self, cache_item: "BaseCacheItem") -> "list[BaseCacheItem]":
        if self._item is None:
            self._item = list()
        self._item.append(cache_item)



from bclib.context import Context
from ..predicate.predicate import Predicate
from typing import Any


class NotEqual (Predicate):
    """Create not equality cheking predicate"""

    def __init__(self, expression: str, value: Any) -> None:
        super().__init__(expression)
        self.__value = value

    async def check_async(self, context: Context) -> bool:
        try:
            value = eval(self.exprossion, {}, {"context": context})
            return self.__value != value
        except:  # pylint: disable=bare-except
            return False



from bclib.cache.cache_status import CacheStatus
from ..cache.signal_base_cache_manager import SignalBaseCacheManager
from bclib.utility import DictEx
from ..cache.cache_status import CacheStatus
from typing import Callable
from ..cache.cache_item.base_cache_item import BaseCacheItem
from ..cache.cache_item.function_cache_item import FunctionCacheItem
from .cache_item.scalar_cache_item import ScalarCacheItem
from ..cache.value_item.base_value_item import BaseValueItem
from ..cache.value_item.array_value_item import ArrayValueItem
from ..cache.value_item.scalar_value_item import ScalarValueItem
from functools import wraps

class InMemoryCacheManager(SignalBaseCacheManager):
    
    def __init__(self, options: DictEx) -> None:
        super().__init__(options)
        self.__cache_dict:"dict[str, BaseValueItem]" = dict()

    def __add_or_update(self, key:"str", cache_item:"BaseCacheItem", value_item:"BaseValueItem") -> "CacheStatus":
        if key not in self.__cache_dict:
            self.__cache_dict[key] = value_item(cache_item)
            return CacheStatus.ADDED
        else:
            try:
                self.__cache_dict[key].add_or_update_item(cache_item)
                return CacheStatus.UPDATED
            except TypeError as ex:
                print(repr(ex))
                return CacheStatus.ERROR

    def cache_decorator(self, key:"str"=None, life_time:"int"=0) -> "Callable":
        """
        Decorator that caches the result of a function for a specified key and life time.

        Args:
            key (str): The key to use for caching the function's result.
            life_time (int): The life time of the cache item in seconds. If not specified, uses the default life time.

        Returns:
            Callable: The decorated function.

        Raises:
            ValueError: If the key is None.
        """
        def decorator(function):
            cache_item = FunctionCacheItem(None, life_time, function)
            function.cache = cache_item
            if key is not None:
                self.__add_or_update(key, cache_item, ArrayValueItem)

            @wraps(function)
            def wrapper(*args, **kwargs):
                function_cache:"FunctionCacheItem" = function.cache
                return function_cache.get_data(*args, **kwargs)
            return wrapper
        return decorator

    def reset(self, keys:"list[str]"=None) -> "CacheStatus":
        """
        Resets the cache by removing all items or specified keys.

        Args:
            keys (list[str]): Optional list of keys to remove from the cache. If not specified, removes all items.

        Returns:
            CacheStatus: The status of the cache after the reset operation.

        Raises:
            None
        """
        if keys is None or len(keys) == 0:
            keys = list(self.__cache_dict.keys())
        for key in keys:
            self.__cache_dict[key].reset()
        return CacheStatus.RESET

    def clean(self) -> "CacheStatus":
        """
        Removes expired cache items from the cache dictionary.
        """
        cleaned_cache_dict = dict()
        for key, value in self.__cache_dict.items():
            if value.get_item() is not None:
                cleaned_cache_dict[key] = value
        self.__cache_dict = cleaned_cache_dict
        return CacheStatus.CLEANED

    def get_cache(self, key:"str") -> "list|any|None":
        return self.__cache_dict[key].get_item() if key in self.__cache_dict else None
    
    def add_or_update(self, key: str, data: "any", life_time:"int"= 0) -> "CacheStatus":
        """
        Add or update an item in the cache.
        Args:
            key (str): The key to identify the cache item.
            data (any): The data to be stored in the cache.
            life_time (int, optional): The time-to-live (TTL) of the cache item in seconds. If 0, the default life time of the cache will be used.

        Returns:
            CacheStatus: The status of the cache after adding or updating the item. Returns CacheStatus.ADDED if a new item was added to the cache, and CacheStatus.UPDATED if an existing item was updated.        
        """
        return self.__add_or_update(key, ScalarCacheItem(data, int(life_time)), ScalarValueItem)



from abc import ABC, abstractmethod
from bclib.utility import DictEx


class SourceContext(ABC):
    """Base class for manage source base context"""

    @property
    @abstractmethod
    def raw_command(self) -> str:
        pass

    @property
    @abstractmethod
    def command(self) -> DictEx:
        pass

    @property
    @abstractmethod
    def params(self) -> DictEx:
        pass

    @property
    @abstractmethod
    def dmn_id(self) -> int:
        pass

    @property
    @abstractmethod
    def process_async(self) -> bool:
        pass



import asyncio

from .message_type import MessageType
from .message import Message


class ReceiveMessage(Message):
    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter,  sessionId: str, messageType: MessageType, buffer: bytes = None) -> None:
        super().__init__(sessionId, messageType, buffer)
        self.reader = reader
        self.writer = writer

    async def read_next_message_async(self) -> 'ReceiveMessage':
        return await ReceiveMessage.read_from_stream_async(self.reader, self.writer)

    @staticmethod
    async def read_from_stream_async(reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> 'ReceiveMessage':
        message: ReceiveMessage = None
        data = await reader.readexactly(1)
        if data:
            message_type = MessageType(int.from_bytes(
                data, byteorder='big', signed=True))
            data = await reader.readexactly(4)
            data_len = int.from_bytes(data, byteorder='big', signed=True)
            data = await reader.readexactly(data_len)
            session_id = data.decode("utf-8")
            parameter = None
            if message_type in (MessageType.AD_HOC, MessageType.MESSAGE, MessageType.CONNECT):
                data = await reader.readexactly(4)
                data_len = int.from_bytes(
                    data, byteorder='big', signed=True)
                data = await reader.readexactly(data_len)
                parameter = data
            message = ReceiveMessage(
                reader, writer, session_id, message_type, parameter)
        return message



from abc import ABC
from ..signaler.base_signaler import BaseSignaler
from bclib.utility import DictEx
from ..signaler.no_signaler import NoSignaler
from ..signaler.rabbit_signaler import RabbitSignaller
from typing import Callable

class SignalerFactory(ABC):
    @staticmethod
    def create(reset_cache_callback:"Callable", signaler_options:"DictEx"=None) -> "BaseSignaler":
        """        
        This is a factory function for create signaler

        Args:
            reset_cache_callback (Callable): The function that reset cache.
            signaler_options (DictEx, optional): The options for signaler type.

        Raises:
            ValueError: If type of signaler inputs in options is not 'rabbit'

        Returns:
            BaseSignaler: BaseSignaler
        """
        if signaler_options is not None and signaler_options.has("type"):
            signaler_type = signaler_options.type
            if signaler_type == "rabbit":
                return RabbitSignaller(reset_cache_callback, signaler_options)
            else:
                raise ValueError(f"Unknown type for signaler ('${signaler_type}')")
        return NoSignaler()


from ..db_manager.db import Db


class SingletonMeta(type):
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `_init_` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class MongoDb(metaclass=SingletonMeta):
    """Mongo implementation of Db wrapper"""

    def __init__(self, connection_string: str) -> None:
        super().__init__()
        import pymongo
        self.client = pymongo.MongoClient(connection_string)

    # def _exit_(self, exc_type, exc_val, exc_tb):
    #     self.client.close()
    #     return super()._exit_(exc_type, exc_val, exc_tb)
